/*
 *
 *  Copyright (c) 2012-2015
 *  name : Francis Banyikwa
 *  email: mhogomchungu@gmail.com
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QWidget>

#include <QObject>
#include <QString>
#include <QFileDialog>
#include <QFile>
#include <QDir>
#include <QAction>
#include <QCloseEvent>

#include "../../zuluCrypt-gui/dialogmsg.h"

namespace Ui {
class MainWindow;
}

class MainWindow : public QWidget
{
	Q_OBJECT
public:
	MainWindow( QWidget * parent = 0 ) ;
	void setToken( char * const * ) ;
	~MainWindow();
private slots:
	void defaultButton( void ) ;
	void pbOpen( void ) ;
	void pbCancel( void ) ;
	void done( void ) ;
private:
	void closeEvent( QCloseEvent * ) ;
	void SetFocus( void );
	Ui::MainWindow * m_ui;
	QString m_token ;
	void * m_handle ;
};

#endif // MAINWINDOW_H
