/*
 *
 *  Copyright (c) 2011-2015
 *  name : Francis Banyikwa
 *  email: mhogomchungu@gmail.com
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "includes.h"

int zuluCryptCloseVolume( const char * map,char ** mount_point )
{
	int i = zuluCryptUnmountVolume( map,mount_point ) ;

	if( i == 0 || i == 3 ){

		if( zuluCryptCloseMapper( map ) == 0 ){

			return 0 ;
		}else{
			return 5 ;
		}
	}else{
		return i ;
	}
}
